module CURIC::Gizmo
  def self.toggle_gizmo
    return unless PLUGIN.active_overlay

    PLUGIN.active_overlay.enabled = !PLUGIN.active_overlay.enabled?
  end

  def self.active_overlay(model = Sketchup.active_model)
    overlay = model.overlays.find { |o| o.is_a?(CURIC::Gizmo::GizmoOverlay) }

    overlay
  end
  
  class GizmoOverlay < Sketchup::Overlay
    include ShapeGeom

    attr_reader :id, :name, :mouse, :flags
    attr_accessor :active_gizmo

    def initialize
      @id = 'Curic.Gizmo.Overlay'
      @name = 'Gizmo'

      @mouse = [0, 0]

      super @id, @name

      start
    end 

    def active_gizmo?
      @active_gizmo
    end

    def active_itself?
      @model.tools.active_tool == self
    end

    def edited?
      @last_transform
    end

    def onMouseMove(flags, x, y, view)
      return if @mouse && @mouse == [x, y]
      
      @flags = flags
      @mouse = [x, y]
      return unless enabled?
      return unless @gizmo

      if active_gizmo?
        # Delay for onUserText
        if active_itself? && edited? && (@edit_mouse && @edit_mouse.distance(@mouse) < 10)
          # Do nothing
        else
          @gizmo.onMouseMove(flags, x, y, view)
        end
      end

      @copy = flags & COPY_MODIFIER_MASK == COPY_MODIFIER_MASK

      at = @model.tools.active_tool
      if at != self
        @model.tools.push_tool(self) if @gizmo.mouse_over?
      else
        @model.tools.pop_tool unless @gizmo.mouse_over?
      end
    rescue StandardError => e
      p e
      @model.tools.pop_tool if active_itself?
    end

    def onLButtonDown(flags, x, y, view)
      @flags = flags
      @mouse = [x, y]
      @gizmo.onLButtonDown(flags, x, y, view) if active_gizmo?
    end

    def onLButtonUp(flags, x, y, view)
      @flags = flags
      @mouse = [x, y]
      @copy = flags & COPY_MODIFIER_MASK == COPY_MODIFIER_MASK

      @gizmo.onLButtonUp(flags, x, y, view) if active_gizmo?
    end
      
    def update_ui
      if active_gizmo?
        hotkey = KEY_MASKS[PLUGIN.settings[:key_mask]]
        status = "Hold #{hotkey} = Pick on list."
        Sketchup.set_status_text status, SB_VCB_LABEL
      end
    end

    def active_bounds
      entities = @model.active_entities
      bb = Geom::BoundingBox.new
      entities.each do |entity|
        bb.add(entity.bounds)
      end
      bb
    end

    def onMouseEnter(*args)
      return unless enabled?
      return if active_itself?

      flags, x, y, _view = args

      @mouse = [x, y]
      @flags = flags
    end

    def onMouseLeave(_view); end
    
    def tool_changed(tool_name)
      return unless enabled?
      
      @active_gizmo = %w[CameraOrbitTool SelectionTool].include?(tool_name)
      @active_gizmo = true if active_itself?
      
      @model.active_view.invalidate
    end

    def selection_changed(_selection)
      update_gizmo
      @model.active_view.invalidate
    end

    def update_gizmo
      if PLUGIN.gizmo_orientation == LOCAL_ORIENTATION && selected_one_object?
        object = @selection[0]
        tr = object.transformation
        bb = object.definition.bounds
        axes = [tr.origin, tr.xaxis, tr.yaxis, tr.zaxis]
      else
        
        bb = Geom::BoundingBox.new
        @model.selection.each do |entity|
          next if entity.respond_to?(:locked?) && entity.locked?
          
          bb.add(entity.bounds)
        end
        tr = IDENTITY
        axes = @model.axes.to_a
      end

      @gizmo.bounds = bb
      @gizmo.transformation = tr
      @gizmo.origin = bb.center.transform(tr)
      @gizmo.xaxis = axes[1]
      @gizmo.yaxis = axes[2]
      @gizmo.zaxis = axes[3]
    end

    def selected_one_object?
      @selection.length == 1 &&
        (@selection[0].is_a?(Sketchup::Group) || @selection[0].is_a?(Sketchup::ComponentInstance))
    end

    def activate
      reset
      start
    end

    # @param [Sketchup::View] view
    def deactivate(_view)
      reset
      stop
      # view.invalidate
    end

    # @param [Sketchup::View] view
    def suspend(view)
      view.invalidate
    end

    # @param [Sketchup::View] view
    def resume(view)
      view.invalidate
    end

    def reset
      @last_transform = nil
      @edit_mouse = nil
      @copy = false
    end

    def selected_entities
      @selection.to_a.find_all { |e| 
        !((e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)) && e.locked?)
      }
    end

    def start
      @model = Sketchup.active_model
      @selection = @model.selection
      @view = @model.active_view

      self.active_gizmo = true
      
      @axes = @model.axes.to_a
      @gizmo = CURIC::Gizmo::Manipulator.new(*@axes)

      @gizmo.on_transform_start do |_action| 
        @model.start_operation('Transform', true)
      end

      @gizmo.on_transform do |transformation, t_total, data|
        next if @selection.empty?

        @last_transform = [transformation, t_total, data]
        @model.active_entities.transform_entities(transformation, selected_entities)
      end

      @gizmo.on_transform_end do |_gizmo|
        if @copy && !selected_entities.empty?
          tr_new = @last_transform[1]
          @model.active_entities.transform_entities(tr_new.inverse, selected_entities)

          temp_group = @model.active_entities.add_group(selected_entities)

          instance = temp_group.copy
          instance.transform!(tr_new)

          ents = instance.explode
          temp_group.explode

          ents.delete_if { |e| e.deleted? || !e.is_a?(Sketchup::Drawingelement) }
          @selection.clear
          @selection.add(ents) unless ents.empty?
        end

        @model.commit_operation
        update_gizmo
        @edit_mouse = @mouse
      end

      @gizmo.on_click do |data|
        action_name, direction = data
        prompts = ['Enter value:']
        value = case action_name
                when :move
                  0.mm
                when :rotate
                  0
                when :scale
                  1.0
                end
        defaults = [value]
        title = action_name.to_s
        title[0] = title[0].upcase
        case action_name
        when :move, :rotate
          title += if direction.parallel?(@gizmo.axes.x.direction)
            ' X'
                   elsif direction.parallel?(@gizmo.axes.y.direction)
            ' Y'
                   else
            ' Z'
                   end
        when :scale
          marks = data[3]
          title += ' X' if marks.x

          title += ' Y' if marks.y

          title += ' Z' if marks.z
        end

        result = UI.inputbox(prompts, defaults, title)
        next unless result

        value = result[0]

        if action_name == :scale && value == 0
          UI.messagebox('Invalid scale')
          next 
        end

        case action_name
        when :move
          vec = direction.clone
          vec.length = value
          tr = Geom::Transformation.translation(vec)
        when :rotate
          tr = Geom::Transformation.rotation(@gizmo.origin, direction, value.degrees)
        when :scale
          marks = data[3]
          scale_x, scale_y, scale_z = scale_marks(marks, value)

          tr = Geom::Transformation.scaling(scale_x, scale_y, scale_z)
          gx, gy, gz = @gizmo.axes.map do |axis|
            axis.direction
          end
          go = @gizmo.origin
          gizmo_tr = Geom::Transformation.new(gx, gy, gz, go)

          tr = gizmo_tr * tr * gizmo_tr.inverse
        end

        @last_transform = [tr, tr, data]
        @model.active_entities.transform_entities(tr, selected_entities)
      end

      update_gizmo
    end

    def scale_marks(marks, value)
      x, y, z = marks
      scale_x = x ? value : 1.0
      scale_y = y ? value : 1.0
      scale_z = z ? value : 1.0
      
      [scale_x, scale_y, scale_z]
    end

    def stop
      # p 'stop'
    end

    def onCancel(reason, view)
      if reason == 0 && @gizmo.active?
        selected = @selection.to_a
        @gizmo.cancel
        @model.abort_operation
        @selection.clear
        @selection.add(selected)
        update_gizmo
        view.invalidate
      end
    end

    def getMenu(menu, _flags, _x, _y, _view)
      ['Global', 'Object'].each_with_index do |name, i|
        cmd = menu.add_item(name) do 
          PLUGIN.gizmo_orientation = i 
          update_gizmo
        end
        menu.set_validation_proc(cmd) { PLUGIN.gizmo_orientation == i ? MF_CHECKED : MF_UNCHECKED }
      end
      menu.add_separator
      menu.add_item('Hide') do 
        PLUGIN.toggle_gizmo 
        Sketchup.active_model.tools.pop_tool
      end

      true
    end

    def onUserText(text, view)
      return unless edited?

      action = @last_transform[2][0]
      case action
      when :move
        value = text.to_l
      when :rotate
        value = text.to_f.degrees
      when :scale
        value = text.to_f
        return UI.messagebox('Invalid scale') if value == 0
      end

      @model.start_operation("Re-edit", true)
      reEdit(value)
      @model.commit_operation

      update_gizmo
      view.invalidate
    rescue ArgumentError
      view.tooltip = 'Invalid length'
    end

    def reEdit(value)
      tr, t_total, data = @last_transform
      action = data[0]
      case action
      when :move
        vec = data[1].clone
        vec.length = value
        tr = Geom::Transformation.translation(vec)
      when :rotate
        tr = Geom::Transformation.rotation(@gizmo.origin, data[2], value)
      when :scale
        marks = data[3]
        scale_x, scale_y, scale_z = scale_marks(marks, value)

        tr = Geom::Transformation.scaling(scale_x, scale_y, scale_z)
        gx, gy, gz = @gizmo.axes.map do |axis|
          axis.direction
        end
        go = @gizmo.origin
        gizmo_tr = Geom::Transformation.new(gx, gy, gz, go)

        tr = gizmo_tr * tr * gizmo_tr.inverse
      end

      @model.active_entities.transform_entities(tr * t_total.inverse, selected_entities)
      t_total = tr
      @last_transform = [tr, t_total, data]
    end

    def getExtents
      bb = @model.bounds
      
      @model.selection.each do |e|
        bb.add(e.bounds)
      end

      bb
    end

    def onKeyDown(key, _repeat, _flags, view)
      if key == COPY_MODIFIER_KEY
        @copy = true
        view.invalidate
      end
    end

    def onKeyUp(key, _repeat, _flags, view)
      if key == COPY_MODIFIER_KEY
        @copy = false
        view.invalidate
      end
    end

    def draw(view)
      return unless active_gizmo? && @gizmo

      @gizmo.draw(view) if @gizmo && @selection.length > 0

      draw_origin_bounds(view) if @gizmo.active? && @copy
    end

    def draw_origin_bounds(view)
      color = inverse_color(@model.rendering_options['HighlightColor'].to_a)
      draw_bounds(view, @gizmo.bounds, @gizmo.transformation, color)
    end

    def inverse_color(rgba)
      r, g, b, a = rgba
    
      inverse_r = 255 - r
      inverse_g = 255 - g
      inverse_b = 255 - b
    
      [inverse_r, inverse_g, inverse_b, a]
    end

    def draw_bounds(view, bb, tr = IDENTITY, color = [255, 0, 0])
      return unless bb
      return if bb.empty?

      pts = 0.upto(7).map { |i| bb.corner(i).transform(tr) }
      view.line_width = 2
      view.line_stipple = ''
      view.drawing_color = color
    
      top = [pts[0], pts[2], pts[3], pts[1]]
      bottom = [pts[4], pts[6], pts[7], pts[5]]
      lines = (top + [top[0]]).each_cons(2).to_a
      lines += (bottom + [bottom[0]]).each_cons(2).to_a
      lines += top.zip(bottom)
      
      view.draw(GL_LINES, lines.flatten) 
    end

    def inspect
      name = self.class.name.split('::').last
      module_name = self.class.name.split('::')[-2]
      hex_id = format('0x%x', (object_id << 1))
      "#<#{module_name}::#{name}:#{hex_id}>"
    end
  end
end
